from matplotlib import pyplot as plt

from detectron2.utils.logger import setup_logger
setup_logger()

# import some common libraries
import numpy as np
import os, json, random
import wandb
import cv2

# import some common detectron2 utilities
from detectron2 import model_zoo
from detectron2.engine import DefaultPredictor
from detectron2.config import get_cfg
from detectron2.utils.visualizer import Visualizer
from detectron2.data import MetadataCatalog, DatasetCatalog

#Importando minhas bibliotecas
from PIL import Image
import glob
import numpy as np
from skimage import io
from skimage.color import rgb2gray
from sklearn.cluster import KMeans
from skimage.transform import rotate
from numpy import asarray
import kwcoco
import numpy
from kwcoco.coco_dataset import *
from PIL import Image
#Metricas
from sklearn.metrics import accuracy_score

#Parte do Treino do Detectron2
from detectron2.engine import DefaultTrainer
from detectron2.data.datasets import register_coco_instances
#import the COCO Evaluator to use the COCO Metrics
from detectron2.evaluation import COCOEvaluator, inference_on_dataset
from detectron2.data import build_detection_test_loader

from detectron2.engine.hooks import HookBase
from detectron2.evaluation import inference_context
from detectron2.utils.logger import log_every_n_seconds
from detectron2.data import DatasetMapper, build_detection_test_loader
import detectron2.utils.comm as comm
import torch
import time
import datetime
import logging

import argparse

parser = argparse.ArgumentParser(description='A test program.')

parser.add_argument("-m", "--modelo", help="Nome do Modelo do Model zoo, exemplo = faster_rcnn_R_50_FPN_1x.yaml")
parser.add_argument("-ann", "--anotations", help="Path das Anotacoes", type=str)
parser.add_argument("-n", "--nome", help="Nome Projeto Wandb", type=str)
# parser.add_argument("-img", "--images", help="Path das Imagens")

args = parser.parse_args()
print(args.anotations)

# Chave do Wandb 13bb9ff285b316aff58ecd7fa677b0225b09cab5

# !wandb login "13bb9ff285b316aff58ecd7fa677b0225b09cab5"

class MyTrainer(DefaultTrainer):
    @classmethod
    def build_evaluator(cls, cfg, dataset_name, output_folder=None):
        if output_folder is None:
            output_folder = os.path.join(cfg.OUTPUT_DIR, "inference")
        return COCOEvaluator(dataset_name, cfg, True, output_folder)
                     
    def build_hooks(self):
        hooks = super().build_hooks()
        hooks.insert(-1,LossEvalHook(
            cfg.TEST.EVAL_PERIOD,
            self.model,
            build_detection_test_loader(
                self.cfg,
                self.cfg.DATASETS.TEST[0],
                DatasetMapper(self.cfg,True)
            )
        ))
        return hooks

class LossEvalHook(HookBase):
    def __init__(self, eval_period, model, data_loader):
        self._model = model
        self._period = eval_period
        self._data_loader = data_loader
    
    def _do_loss_eval(self):
        # Copying inference_on_dataset from evaluator.py
        total = len(self._data_loader)
        num_warmup = min(5, total - 1)
            
        start_time = time.perf_counter()
        total_compute_time = 0
        losses = []
        for idx, inputs in enumerate(self._data_loader):            
            if idx == num_warmup:
                start_time = time.perf_counter()
                total_compute_time = 0
            start_compute_time = time.perf_counter()
            if torch.cuda.is_available():
                torch.cuda.synchronize()
            total_compute_time += time.perf_counter() - start_compute_time
            iters_after_start = idx + 1 - num_warmup * int(idx >= num_warmup)
            seconds_per_img = total_compute_time / iters_after_start
            if idx >= num_warmup * 2 or seconds_per_img > 5:
                total_seconds_per_img = (time.perf_counter() - start_time) / iters_after_start
                eta = datetime.timedelta(seconds=int(total_seconds_per_img * (total - idx - 1)))
                log_every_n_seconds(
                    logging.INFO,
                    "Loss on Validation  done {}/{}. {:.4f} s / img. ETA={}".format(
                        idx + 1, total, seconds_per_img, str(eta)
                    ),
                    n=5,
                )
            loss_batch = self._get_loss(inputs)
            losses.append(loss_batch)
        mean_loss = np.mean(losses)
        self.trainer.storage.put_scalar('validation_loss', mean_loss)
        comm.synchronize()

        return losses
            
    def _get_loss(self, data):
        # How loss is calculated on train_loop 
        metrics_dict = self._model(data)
        metrics_dict = {
            k: v.detach().cpu().item() if isinstance(v, torch.Tensor) else float(v)
            for k, v in metrics_dict.items()
        }
        total_losses_reduced = sum(loss for loss in metrics_dict.values())
        return total_losses_reduced
        

    def after_step(self):
        next_iter = self.trainer.iter + 1
        is_final = next_iter == self.trainer.max_iter
        if is_final or (self._period > 0 and next_iter % self._period == 0):
            self._do_loss_eval()
        self.trainer.storage.put_scalars(timetest=12)


configName = []
configName.append(args.modelo)

contadorConfig = 1

for Cnome in configName:
  for contFold in range(4):
    train = "gli_train" + str(contadorConfig)
    val = "gli_val" + str(contadorConfig)

    pathCocoTrain = args.anotations + "/Dataset/CrossValidation" + "/train" + str(contFold) + ".json" 
    pathCocoTest = args.anotations + "/Dataset/CrossValidation" + "/test" + str(contFold) + ".json"
    pathImages = args.anotations  + "/Dataset/images"

    register_coco_instances(train, {}, pathCocoTrain, pathImages)

    register_coco_instances(val, {}, pathCocoTest, pathImages)

    wandb.init(project= args.nome, entity="nishinotsk",sync_tensorboard=True,config={"Nome do Modelo": Cnome,"Fold": "train" + str(contFold)})

    cfg = get_cfg()
    cfg.merge_from_file(model_zoo.get_config_file("COCO-Detection/" + Cnome))

    cfg.DATASETS.TRAIN = (train,)
    cfg.DATASETS.TEST = (val,)
    cfg.DATALOADER.NUM_WORKERS = 2
    cfg.SOLVER.IMS_PER_BATCH = 2 #Batch Size
    cfg.SOLVER.BASE_LR = 0.00025 # 0.0025 ou 0.000025
    cfg.TEST.EVAL_PERIOD = 3000
    cfg.SOLVER.WARMUP_ITERS = 0
    cfg.SOLVER.MAX_ITER = 29144
    cfg.SOLVER.CHECKPOINT_PERIOD = 10000
    cfg.INPUT.MIN_SIZE_TRAIN = 800
    cfg.INPUT.MAX_SIZE_TRAIN = 1333
    cfg.SEED = 5
    cfg.INPUT.RANDOM_FLIP  = "none"
    cfg.TEST.AUG.FLIP = False
    cfg.OUTPUT_DIR = Cnome.split('.')[0] + str(contFold)

    cfg.MODEL.ROI_HEADS.BATCH_SIZE_PER_IMAGE = 256
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = 10  # Número de classes

    os.makedirs(cfg.OUTPUT_DIR, exist_ok=True)
    trainer = MyTrainer(cfg)
    trainer.resume_or_load(resume=False)
    trainer.train()

    cfg = get_cfg()
    cfg.OUTPUT_DIR = Cnome.split('.')[0] + str(contFold)
    cfg.merge_from_file(model_zoo.get_config_file("COCO-Detection/" + Cnome))
    cfg.MODEL.DEVICE = "cuda"
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = 10
    cfg.INPUT.MIN_SIZE_TRAIN = 800
    cfg.INPUT.MAX_SIZE_TRAIN = 1333
    cfg.SEED = 5
    cfg.INPUT.RANDOM_FLIP  = "none"
    cfg.TEST.AUG.FLIP = False
    cfg.TEST.DETECTIONS_PER_IMAGE = 10
    cfg.merge_from_list(["MODEL.WEIGHTS", cfg.OUTPUT_DIR + "/" + "model_final.pth"])
    predictor = DefaultPredictor(cfg)

  # Pós Processamento por KMEANS ----------------------------------------------------------
    self = CocoDataset(pathCocoTest)

    gids = list(self.imgs.keys())
    images = self.images(gids)

    imagensNome = images.lookup('file_name')

    limiar = 0.6

    y_pred = []

    cont = 0
    contImagem = 0
    for i in range(len(imagensNome)):
      im = cv2.imread(pathImages + "/" + imagensNome[contImagem])
      contImagem += 1
      outputs = predictor(im)
      condicao = True

      v = Visualizer(im[:,:,::-1], MetadataCatalog.get(cfg.DATASETS.TRAIN[0]), scale=1.0)
      # out = v.draw_instance_predictions(outputs["instances"].to("cpu"))
      output_pred_boxes = outputs["instances"][outputs['instances'].scores > limiar].pred_boxes.tensor.cpu().numpy()
      classes = outputs["instances"][outputs['instances'].scores > limiar].pred_classes.to("cpu").numpy()
      auxScores = outputs['instances'].scores > limiar
      scores = numpy.zeros(len(classes))
      # ---------------------------------------------------------------------------------------------------------------
      cont = 0
      for item in auxScores:
        if item == True:
          scores[cont] = outputs['instances'].scores[cont]
          cont += 1
      out = v.draw_instance_predictions(outputs["instances"][outputs['instances'].scores > limiar].to("cpu"))
    # ---------------------------------------------------------------------------------------------------------------
      if(len(classes) > 0):
        classes = classes.tolist()
        scores = scores.tolist()
        for item in output_pred_boxes:
          if item[0] == 0:
            condicao = False

        if len(classes) > 3 and condicao == True:
          qtd = len(output_pred_boxes)
          vetor = np.zeros((qtd,2))
          cont = 0
          for item in range(len(output_pred_boxes)):
            vetor[item][cont] = output_pred_boxes[item][0]
            vetor[item][cont+1] = output_pred_boxes[item][2]

          vetor = vetor.tolist()
          #Parte do Kmeas ----------------------------------------------------------------------------------------------
          kmeans = KMeans(n_clusters=3, random_state=0).fit(vetor)
          kcls = kmeans.predict(vetor)
          #-------------------------------------------------------------------------------------------------------------
          matriz = np.zeros((len(output_pred_boxes),len(output_pred_boxes)))
          vetorScores = np.zeros(len(output_pred_boxes))
          cont = np.zeros(3)
          cont = cont.astype(int)
          for a in range(len(output_pred_boxes)):
            matriz[kcls[a]][cont[kcls[a]]] = output_pred_boxes[a][0]
            vetorScores[a] = (kcls[a] * 10) + cont[kcls[a]]
            cont[kcls[a]] = cont[kcls[a]] + 1

          matriz = matriz.tolist()
          vetorScores = vetorScores.tolist()
          # break
          #-------------------------------------------------------------------------------------------------------------
          calculoDoScore = np.zeros(3)
          indexDoScore = np.zeros(3)
          for i in range(3):
            for j in range(4):
              item = matriz[i][j]
              if item != 0:
                index = vetorScores.index((i * 10) + j)
                if scores[index] > calculoDoScore[i]:
                    calculoDoScore[i] = scores[index]
                    indexDoScore[i] = (i * 10) + j
          #-------------------------------------------------------------------------------------------------------------
          left = np.zeros(3)
          cont = 0
          for item in indexDoScore:
            i = int(item / 10)
            j = int(item % 10)
            left[cont] = matriz[i][j]
            cont += 1
          leftLista = left.tolist()
          leftSorted = sorted(left)
          #-------------------------------------------------------------------------------------------------------------
          classKM = np.zeros(3)
          for i in range(3):
            index = scores.index(calculoDoScore[i])
            classKM[i] = classes[index]
          classKM = classKM.astype(numpy.int64)
          #-------------------------------------------------------------------------------------------------------------
          numFinal = np.zeros(3)
          for i in range(len(leftSorted)):
            index = leftLista.index(leftSorted[i])
            numFinal[i] = classKM[index]
          numFinal = numFinal.astype(numpy.int64)
          numFinal = numFinal.tolist()
          numFinal = list(map(int, numFinal))
          a_str = ''.join(str(x) for x in numFinal)

          y_pred.append(a_str)
          # print(numFinal)
        else:
          qtd = len(output_pred_boxes)
          vetor = np.zeros(qtd)

          for item in range(len(output_pred_boxes)):
            vetor[item] = output_pred_boxes[item][0]

          aux = sorted(vetor)
          vetor = list(vetor)
          classFinal = np.zeros(qtd)
          contador = 0

          for item in aux:
            classFinal[contador] = classes[vetor.index(item)]
            contador += 1

          classFinal = list(map(int, classFinal))
          a_str = ''.join(str(x) for x in classFinal)
          
          y_pred.append(a_str)
      else:
        y_pred.append("-1")

    cont = 0
    for item in y_pred:
      y_pred[cont] = int(y_pred[cont])
      cont += 1

    # YREAL

    self = CocoDataset(pathCocoTest)

    gids = list(self.imgs.keys())
    images = self.images(gids)
    y_true = []
    cont = 0

    for id in images:
      AllBoxes = []
      #Pegando os dados da anotação
      categoryid = images.annots.lookup('category_id')[cont]
      lbox = images.annots.lookup('bbox')[cont]
      # print(lbox[cont])
      cont = cont + 1

      #Adicionar no Dataset
      for i in range(len(lbox)):
        cid = categoryid[i]
        bbox = lbox[i][0]
        AllBoxes.append(bbox)

      leftSorted = sorted(AllBoxes)
      finalClass = []
      for item in leftSorted:
        index = AllBoxes.index(item)
        finalClass.append(categoryid[index]-1)

      finalClass = list(map(int, finalClass))
      a_str = ''.join(str(x) for x in finalClass)
      
      y_true.append(a_str)

    cont = 0
    for item in y_true:
      y_true[cont] = int(y_true[cont])
      cont += 1

    # Cálculo da Acurácia
    acc = accuracy_score(y_true, y_pred)

    wandb.config.update({"acuracia KMEANS": acc})

  # Pós Processamento por Posição ----------------------------------------------------------
    self = CocoDataset(pathCocoTest)

    gids = list(self.imgs.keys())
    images = self.images(gids)

    imagensNome = images.lookup('file_name')

    y_pred = []

    cont = 0
    contImagem = 0
    path = pathImages + "/"
    for item in imagensNome:
      im = cv2.imread(path + item)
      contImagem += 1
      outputs = predictor(im)
      
      limiar = 0.6

      v = Visualizer(im[:,:,::-1], MetadataCatalog.get(cfg.DATASETS.TRAIN[0]), scale=1.0)
      output_pred_boxes = outputs["instances"][outputs['instances'].scores > limiar].pred_boxes.tensor.cpu().numpy()
      classes = outputs["instances"][outputs['instances'].scores > limiar].pred_classes.to("cpu").numpy()
      out = v.draw_instance_predictions(outputs["instances"][outputs['instances'].scores > limiar].to("cpu"))
    # ---------------------------------------------------------------------------------------------------------------

      qtd = len(output_pred_boxes)
      vetor = np.zeros(qtd)
      if qtd > 0:
        for item in range(len(output_pred_boxes)):
          vetor[item] = output_pred_boxes[item][0]

        aux = sorted(vetor)
        vetor = list(vetor)
        classFinal = np.zeros(qtd)
        contador = 0

        for item in aux:
          classFinal[contador] = classes[vetor.index(item)]
          contador += 1

        classFinal = list(map(int, classFinal))
        a_str = ''.join(str(x) for x in classFinal)
        
        y_pred.append(a_str)
        # yDigitos[cont] = float(a_str.replace(',', '.'))
        # cont += 1
      else:
        y_pred.append("-1")

    cont = 0
    for item in y_pred:
      y_pred[cont] = int(y_pred[cont])
      cont += 1

    # Cálculo da Acurácia
    acc = accuracy_score(y_true, y_pred)


    wandb.config.update({"acuracia Posição": acc})

  # Pós Processamento por NMS 0.3 ----------------------------------------------------------

    cfg = get_cfg()
    cfg.OUTPUT_DIR = Cnome.split('.')[0] + str(contFold)
    cfg.merge_from_file(model_zoo.get_config_file("COCO-Detection/" + Cnome))
    cfg.MODEL.DEVICE = "cuda"
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = 10
    cfg.INPUT.MIN_SIZE_TRAIN = 800
    cfg.INPUT.MAX_SIZE_TRAIN = 1333
    cfg.SEED = 5
    cfg.INPUT.RANDOM_FLIP  = "none"
    cfg.TEST.AUG.FLIP = False
    cfg.MODEL.ROI_HEADS.NMS_THRESH_TEST = 0.3
    cfg.TEST.DETECTIONS_PER_IMAGE = 10
    cfg.merge_from_list(["MODEL.WEIGHTS", cfg.OUTPUT_DIR + "/" + "model_final.pth"])
    predictor = DefaultPredictor(cfg)

    self = CocoDataset(pathCocoTest)

    gids = list(self.imgs.keys())
    images = self.images(gids)

    imagensNome = images.lookup('file_name')

    y_pred = []

    cont = 0
    contImagem = 0
    path = pathImages + "/"
    for item in imagensNome:
      im = cv2.imread(path + item)
      contImagem += 1
      outputs = predictor(im)
      
      limiar = 0.6

      v = Visualizer(im[:,:,::-1], MetadataCatalog.get(cfg.DATASETS.TRAIN[0]), scale=1.0)
      output_pred_boxes = outputs["instances"][outputs['instances'].scores > limiar].pred_boxes.tensor.cpu().numpy()
      classes = outputs["instances"][outputs['instances'].scores > limiar].pred_classes.to("cpu").numpy()
      out = v.draw_instance_predictions(outputs["instances"][outputs['instances'].scores > limiar].to("cpu"))
    # ---------------------------------------------------------------------------------------------------------------

      qtd = len(output_pred_boxes)
      vetor = np.zeros(qtd)
      if qtd > 0:
        for item in range(len(output_pred_boxes)):
          vetor[item] = output_pred_boxes[item][0]

        aux = sorted(vetor)
        vetor = list(vetor)
        classFinal = np.zeros(qtd)
        contador = 0

        for item in aux:
          classFinal[contador] = classes[vetor.index(item)]
          contador += 1

        classFinal = list(map(int, classFinal))
        a_str = ''.join(str(x) for x in classFinal)
        
        y_pred.append(a_str)
        # yDigitos[cont] = float(a_str.replace(',', '.'))
        # cont += 1
      else:
        y_pred.append("-1")

    cont = 0
    for item in y_pred:
      y_pred[cont] = int(y_pred[cont])
      cont += 1

    # Cálculo da Acurácia
    acc = accuracy_score(y_true, y_pred)


    wandb.config.update({"acuracia NMS 0.3": acc})

  # Pós Processamento por NMS 0.4 ----------------------------------------------------------

    cfg = get_cfg()
    cfg.OUTPUT_DIR = Cnome.split('.')[0] + str(contFold)
    cfg.merge_from_file(model_zoo.get_config_file("COCO-Detection/" + Cnome))
    cfg.MODEL.DEVICE = "cuda"
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = 10
    cfg.INPUT.MIN_SIZE_TRAIN = 800
    cfg.INPUT.MAX_SIZE_TRAIN = 1333
    cfg.SEED = 5
    cfg.INPUT.RANDOM_FLIP  = "none"
    cfg.TEST.AUG.FLIP = False
    cfg.TEST.DETECTIONS_PER_IMAGE = 10
    cfg.MODEL.ROI_HEADS.NMS_THRESH_TEST = 0.4
    cfg.merge_from_list(["MODEL.WEIGHTS", cfg.OUTPUT_DIR + "/" + "model_final.pth"])
    predictor = DefaultPredictor(cfg)

    self = CocoDataset(pathCocoTest)

    gids = list(self.imgs.keys())
    images = self.images(gids)

    imagensNome = images.lookup('file_name')

    y_pred = []

    cont = 0
    contImagem = 0
    path = pathImages + "/"
    for item in imagensNome:
      im = cv2.imread(path + item)
      contImagem += 1
      outputs = predictor(im)
      
      limiar = 0.6

      v = Visualizer(im[:,:,::-1], MetadataCatalog.get(cfg.DATASETS.TRAIN[0]), scale=1.0)
      output_pred_boxes = outputs["instances"][outputs['instances'].scores > limiar].pred_boxes.tensor.cpu().numpy()
      classes = outputs["instances"][outputs['instances'].scores > limiar].pred_classes.to("cpu").numpy()
      out = v.draw_instance_predictions(outputs["instances"][outputs['instances'].scores > limiar].to("cpu"))
    # ---------------------------------------------------------------------------------------------------------------

      qtd = len(output_pred_boxes)
      vetor = np.zeros(qtd)
      if qtd > 0:
        for item in range(len(output_pred_boxes)):
          vetor[item] = output_pred_boxes[item][0]

        aux = sorted(vetor)
        vetor = list(vetor)
        classFinal = np.zeros(qtd)
        contador = 0

        for item in aux:
          classFinal[contador] = classes[vetor.index(item)]
          contador += 1

        classFinal = list(map(int, classFinal))
        a_str = ''.join(str(x) for x in classFinal)
        
        y_pred.append(a_str)
        # yDigitos[cont] = float(a_str.replace(',', '.'))
        # cont += 1
      else:
        y_pred.append("-1")

    cont = 0
    for item in y_pred:
      y_pred[cont] = int(y_pred[cont])
      cont += 1

    # Cálculo da Acurácia
    acc = accuracy_score(y_true, y_pred)

    wandb.config.update({"acuracia NMS 0.4": acc})

  # Pós Processamento por NMS 0.6 ----------------------------------------------------------

    cfg = get_cfg()
    cfg.OUTPUT_DIR = Cnome.split('.')[0] + str(contFold)
    cfg.merge_from_file(model_zoo.get_config_file("COCO-Detection/" + Cnome))
    cfg.MODEL.DEVICE = "cuda"
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = 10
    cfg.INPUT.MIN_SIZE_TRAIN = 800
    cfg.INPUT.MAX_SIZE_TRAIN = 1333
    cfg.SEED = 5
    cfg.INPUT.RANDOM_FLIP  = "none"
    cfg.TEST.AUG.FLIP = False
    cfg.MODEL.ROI_HEADS.NMS_THRESH_TEST = 0.6
    cfg.TEST.DETECTIONS_PER_IMAGE = 10
    cfg.merge_from_list(["MODEL.WEIGHTS", cfg.OUTPUT_DIR + "/" + "model_final.pth"])
    predictor = DefaultPredictor(cfg)

    self = CocoDataset(pathCocoTest)

    gids = list(self.imgs.keys())
    images = self.images(gids)

    imagensNome = images.lookup('file_name')

    y_pred = []

    cont = 0
    contImagem = 0
    path = pathImages + "/"
    for item in imagensNome:
      im = cv2.imread(path + item)
      contImagem += 1
      outputs = predictor(im)
      
      limiar = 0.6

      v = Visualizer(im[:,:,::-1], MetadataCatalog.get(cfg.DATASETS.TRAIN[0]), scale=1.0)
      output_pred_boxes = outputs["instances"][outputs['instances'].scores > limiar].pred_boxes.tensor.cpu().numpy()
      classes = outputs["instances"][outputs['instances'].scores > limiar].pred_classes.to("cpu").numpy()
      out = v.draw_instance_predictions(outputs["instances"][outputs['instances'].scores > limiar].to("cpu"))
    # ---------------------------------------------------------------------------------------------------------------

      qtd = len(output_pred_boxes)
      vetor = np.zeros(qtd)
      if qtd > 0:
        for item in range(len(output_pred_boxes)):
          vetor[item] = output_pred_boxes[item][0]

        aux = sorted(vetor)
        vetor = list(vetor)
        classFinal = np.zeros(qtd)
        contador = 0

        for item in aux:
          classFinal[contador] = classes[vetor.index(item)]
          contador += 1

        classFinal = list(map(int, classFinal))
        a_str = ''.join(str(x) for x in classFinal)
        
        y_pred.append(a_str)
        # yDigitos[cont] = float(a_str.replace(',', '.'))
        # cont += 1
      else:
        y_pred.append("-1")

    cont = 0
    for item in y_pred:
      y_pred[cont] = int(y_pred[cont])
      cont += 1

    # Cálculo da Acurácia
    acc = accuracy_score(y_true, y_pred)

    contadorConfig +=1

    wandb.config.update({"acuracia NMS 0.6": acc})

    wandb.finish()